# Name:       SubStanding for Assetto Corsa
# Version:    v1.4
# Author:     Sylvain Villet, leBluem
# Contact:    sylvain.villet@gmail.com
# Date:       25.04.2020
# Desc.:      This app provides standings and positions
#             in a subgroup of the cars, that can be
#             the same car as the player or a class like
#             LMP1, GT2, GT3 etc defined here.
#             It also provides the global standing with
#             the same design (max 36 cars).
#             The app can be fully configured from the SubStanding_config widget.
#
#             Feel free to add, remove or modify the "classes"
#             definitions in the Substanding_classes/classes.ini file.
#
# Thanks:     - Rombik for the sim_info module
#             - Rivali (OV1Info) and Fernando Deutsch (ferito-LiveCarTracker)
#             for the inspiration and example
#             - ExOAte for the beta tests and ideas
#             See also changelog.txt for details. 

# Global default values, must match the ones in "settings_defaults.ini"
updateTime = 200
bdone=False

# Standings
showLogo = False
showTitle = False
shortNames = True
showBorder = False
showBadge = True
bUseMiniPreview = True # only with badge on
showTyres = True
showPitStops = True
showDelta = True
fontSize = 16
opacity = 20
colorAt = 400
unit = "time"
driverDisplayedGlobal = 3
driverDisplayedClass = 3
driverDisplayedSameCar = 3

# Positions
fontSizePos = 32
opacityPos = 20
showLogoPos = False
showTitlePos = True
showBorderPos = False

import ac, acsys
import sys, os, platform, time
import traceback, functools
import random, configparser, io, re, string

# import libraries
if platform.architecture()[0] == "64bit":
    sysdir='apps/python/SubStanding/SubStanding_dll/stdlib64'
else:
    sysdir='apps/python/SubStanding/SubStanding_dll/stdlib'
sys.path.insert(0, sysdir)
os.environ['PATH'] = os.environ['PATH'] + ";."
import ctypes.wintypes
from SubStanding_lib.sim_info import info

userDir=''
CSIDL_PERSONAL = 5 # My Documents
SHGFP_TYPE_CURRENT = 0 # Get current, not default value
buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
ctypes.windll.shell32.SHGetFolderPathW(None, CSIDL_PERSONAL, None, SHGFP_TYPE_CURRENT, buf)
userDir = buf.value # bingo, userdir/documents!
settingsfile = "apps/python/SubStanding/settings/settings.ini" # when empty, all the default values will stay as above
racelogpath = userDir + "\\Assetto Corsa\\logs\\log.txt"

# Classes from the classes.ini file
classes = []

# Teammates and rivals from the drivers.ini file
teammates = []
rivals = []

# Global variables
serverIp = ""
driversFullList = []
badge=[]
miniskin=[]
myClass = ["No class"]

spacing = 5
firstSpacing = 30
fontSizeConfig = 16
maxDriverCount = 32
myDriver = 0
myCarName = 0
trackLength = 0
lastUpdateTime = 0
debug = 0
cpuUsage = 0

# Objects
SubStandings = {
    "model": 0,
    "class": 0,
    "global": 0
}

SubPositions = {
    "model": 0,
    "class": 0,
    "global": 0
}
config = 0
classesConfig = 0
configApp = 0

def acMain(ac_version):
    try:
        global myCarName, myClass, trackLength, cpuUsage
        global config, classesConfig, configApp
        global updateTime
        global showLogo, showTitle, fontSize, opacity, showBorder, showDelta, colorAt, unit
        global driverDisplayedGlobal, driverDisplayedClass, driverDisplayedSameCar
        global showLogoPos, showTitlePos, fontSizePos, opacityPos, showBorderPos
        global badge, miniskin, showBadge, showTyres, showPitStops
        global teammates, rivals, maxDriverCount
        global myDriver, driversFullList
        global racelogpath, bUseMiniPreview, settingsfile, shortNames

        ac.initFont(0, "Segoe UI", 0, 0)
        ac.initFont(0, "Segoe UI; Weight=Light", 0, 0)
        ac.initFont(0, "Segoe UI Light", 0, 0)
        ac.initFont(0, "Segoe UI Black", 0, 0)

        # - config parser not compatible with CM comments -
        #config = configparser.ConfigParser()
        #config.read()

        if not os.path.isfile(settingsfile):
            with io.open(settingsfile, 'w', encoding='utf-8') as sf:
                sf.close
                ac.log("SubStanding: Settings file created")

        # Global
        updateTime = int(appReadCFGValue(settingsfile, "GLOBAL", "updateTime", updateTime))

        # Standings
        if appReadCFGValue(settingsfile, "STANDINGS", "showLogo", str(showLogo))=="True":
            showLogo=True
        else:
            showLogo=False
        if appReadCFGValue(settingsfile, "STANDINGS", "showTitle", str(showTitle))=="True":
            showTitle=True
        else:
            showTitle=False
        if appReadCFGValue(settingsfile, "STANDINGS", "shortNames", str(shortNames))=="False":
            shortNames = False
        else:
            shortNames = True
        if appReadCFGValue(settingsfile, "STANDINGS", "showBorder", str(showBorder))=="True":
            showBorder = True
        else:
            showBorder = False
        if appReadCFGValue(settingsfile, "STANDINGS", "showBadge", str(showBadge))=="False":
            showBadge = False
        else:
            showBadge = True
        if appReadCFGValue(settingsfile, "STANDINGS", "UseMiniPreview", str(bUseMiniPreview))=="False":
            bUseMiniPreview = False
        else:
            bUseMiniPreview = True
        if appReadCFGValue(settingsfile, "STANDINGS", "showTyres", str(showTyres))=="False":
            showTyres = False
        else:
            showTyres = True
        if appReadCFGValue(settingsfile, "STANDINGS", "showPitStops", str(showPitStops))=="False":
            showPitStops = False
        else:
            showPitStops = True
        if appReadCFGValue(settingsfile, "STANDINGS", "showDelta", str(showDelta))=="False":
            showDelta = False
        else:
            showDelta = True

        fontSize  = int(appReadCFGValue(settingsfile, "STANDINGS", "fontSize", fontSize))
        opacity   = int(appReadCFGValue(settingsfile, "STANDINGS", "opacity", opacity))
        colorAt   = int(appReadCFGValue(settingsfile, "STANDINGS", "colorAt", colorAt))
        unit      = str(appReadCFGValue(settingsfile, "STANDINGS", "unit", unit))
        driverDisplayedGlobal  = int(appReadCFGValue(settingsfile, "STANDINGS", "driverDisplayedGlobal", driverDisplayedGlobal))
        driverDisplayedClass   = int(appReadCFGValue(settingsfile, "STANDINGS", "driverDisplayedClass", driverDisplayedClass))
        driverDisplayedSameCar = int(appReadCFGValue(settingsfile, "STANDINGS", "driverDisplayedSameCar", driverDisplayedSameCar))

        # Positions
        if appReadCFGValue(settingsfile,"POSITIONS", "showLogoPos", str(showLogoPos))=="True":
            showLogoPos  = True
        else:
            showLogoPos  = False
        if appReadCFGValue(settingsfile,"POSITIONS", "showTitlePos", str(showTitlePos))=="True":
            showTitlePos = True
        else:
            showTitlePos = False
        if appReadCFGValue(settingsfile,"POSITIONS", "showBorderPos", str(showBorderPos))=="True":
            showBorderPos = True
        else:
            showBorderPos = False

        fontSizePos  = int(appReadCFGValue(settingsfile,"POSITIONS", "fontSizePos", fontSizePos))
        opacityPos   = int(appReadCFGValue(settingsfile,"POSITIONS", "opacityPos", opacityPos))

        # Get classes
        classesConfig = configparser.ConfigParser()
        classesConfig.read("apps/python/SubStanding/SubStanding_classes/classes.ini")

        for eachSection in classesConfig.sections():
            classTmp = []
            classTmp.append(eachSection)
            for eachItem in classesConfig.items(eachSection):
                classTmp.append(eachItem[1])
            classes.append(classTmp)

        myCarName = ac.getCarName(0)

        for eachClass in classes:
            if myCarName in eachClass:
                myClass = eachClass
                break

        maxDriverCount = ac.getCarsCount()

        # Get badges
        for i in range(maxDriverCount):
            carName = ac.getCarName(i)
            if type(carName) == type(""):
                textureId = ac.newTexture("content/cars/" + carName + "/ui/badge.png")
                badge.append(textureId)

                textureId = -1
                # ie \content\cars\gd_austin_100\skins\GreenMatte\preview.jpg
                sSkin = appReadCFGValue(racelogpath, 'CAR_'+str(i), 'SKIN','')
                # ac.log("content/cars/" + carName + "/skins/" + sSkin + "/preview.png")
                if os.path.isfile(            "content/cars/" + carName + "/skins/" + sSkin + "/preview_mini.png"):
                    textureId = ac.newTexture("content/cars/" + carName + "/skins/" + sSkin + "/preview_mini.png")
                elif os.path.isfile(          "content/cars/" + carName + "/skins/" + sSkin + "/preview_mini.jpg"):
                    textureId = ac.newTexture("content/cars/" + carName + "/skins/" + sSkin + "/preview_mini.jpg")
                elif os.path.isfile(          "content/cars/" + carName + "/skins/" + sSkin + "/preview.png"):
                    textureId = ac.newTexture("content/cars/" + carName + "/skins/" + sSkin + "/preview.png")
                elif os.path.isfile(          "content/cars/" + carName + "/skins/" + sSkin + "/preview.jpg"):
                    textureId = ac.newTexture("content/cars/" + carName + "/skins/" + sSkin + "/preview.jpg")
                miniskin.append(textureId)
            else:
                badge.append(-1)
                miniskin.append(-1)

        # Get teammates and rivals
        driversConfig = configparser.ConfigParser()
        driversConfig.read("apps/python/SubStanding/SubStanding_drivers/drivers.ini")

        for eachSection in driversConfig.sections():
            if eachSection == "TEAMMATES":
                for eachItem in driversConfig.items(eachSection):
                    teammates.append(eachItem[1])
            elif eachSection == "RIVALS":
                for eachItem in driversConfig.items(eachSection):
                    rivals.append(eachItem[1])

        SubStandings["model"]  = StandingWindow("SubStanding_SameCar", "Same Car", driverDisplayedSameCar)
        SubStandings["class"]  = StandingWindow("SubStanding_Class", myClass[0], driverDisplayedClass)
        SubStandings["global"] = StandingWindow("SubStanding_Global", "Global", driverDisplayedGlobal)

        SubPositions["model"]  = PositionWindow("SubStanding_SameCar_Pos", "Same Car")
        SubPositions["class"]  = PositionWindow("SubStanding_Class_Pos", myClass[0])
        SubPositions["global"] = PositionWindow("SubStanding_Global_Pos", "Global")

        configApp = SubStanding_config("SubStanding_config", "SubStanding config", fontSizeConfig)
        configApp.updateView()

        ac.addRenderCallback(SubStandings["model"].window, onRenderCallbackStandingModel)
        ac.addRenderCallback(SubStandings["class"].window, onRenderCallbackStandingClass)
        ac.addRenderCallback(SubStandings["global"].window, onRenderCallbackStandingGlobal)

        ac.addRenderCallback(SubPositions["model"].window, onRenderCallbackPositionModel)
        ac.addRenderCallback(SubPositions["class"].window, onRenderCallbackPositionClass)
        ac.addRenderCallback(SubPositions["global"].window, onRenderCallbackPositionGlobal)

        ac.addRenderCallback(configApp.window, onRenderCallbackConfig)

        trackLength = ac.getTrackLength(0)

        # Get all needed infos
        for index in range(maxDriverCount):
            driver = {
                "carId":              index,
                "connected":          ac.isConnected(index),
                "driverName":         ac.getDriverName(index),
                "driverNameShort":    ac.getDriverName(index),
                "aiStrength":         "",
                "carName":            ac.getCarName(index),
                "bestLap":            ac.getCarState(index, acsys.CS.BestLap),
                "lapTime":            ac.getCarState(index, acsys.CS.LapTime),
                "lapCount":           ac.getCarState(index, acsys.CS.LapCount),
                "speedMS":            ac.getCarState(index, acsys.CS.SpeedMS),
                "distance":           ac.getCarState(index, acsys.CS.NormalizedSplinePosition),
                # ac.getCarState      (index, acsys.CS.LapCount) +
                "tyres":              ac.getCarTyreCompound(index),
                "isInPit":            ac.isCarInPit(index),
                "isInPitLine":        ac.isCarInPitline(index),
                "pitStops":           0,
                "pitEntryFlag":       False,
                "raceStartFlag":      True,
                #"realtimePos":        ac.getCarLeaderboardPosition(index),
                "realtimePos":        ac.getCarRealTimeLeaderboardPosition(index),
                # "realtimePos":        ac.getCarState(index, acsys.CS. carRealTimeLeaderboardPosition),
                "state":              0
            }
            # if '(' in driver["driverNameShort"]:
            sp=driver["driverNameShort"].split('(')
            if len(sp)>1:
                # rest=sp[1]
                driver["driverNameShort"]=sp[0].strip()
                driver["driverName"]=driver["driverNameShort"]
                driver["aiStrength"]=sp[1].split(')')[0]

            sn = driver["driverNameShort"].split(' ')
            if len(sn)>1:
                if len(sn[1])<3:
                    driver["driverNameShort"] = str(sn[0][:6]) + '.' + str(sn[1][:2])
                else:
                    driver["driverNameShort"] = str(sn[0][:2]) + '.' + str(sn[1][:6])
            else:
                driver["driverNameShort"] = str(driver["driverNameShort"][:8])
            if index == 0:
                myDriver = driver
            driversFullList.append(driver)
        if debug:
            cpuUsage = CpuUsage("SubStanding_CpuUsage", "CPU Usage", fontSize)

        return "SubStanding"
    except Exception as e:
        ac.log("SubStanding: Error in acMain: %s" % e + traceback.format_exc())


def writeParameters():
    global settingsfile
    try:
        appWriteCFGValue(settingsfile, "GLOBAL", "updateTime", str(int(updateTime)))

        appWriteCFGValue(settingsfile, "STANDINGS", "showLogo", str(showLogo))
        appWriteCFGValue(settingsfile, "STANDINGS", "showTitle", str(showTitle))
        appWriteCFGValue(settingsfile, "STANDINGS", "shortNames", str(shortNames))
        appWriteCFGValue(settingsfile, "STANDINGS", "fontSize", str(fontSize))
        appWriteCFGValue(settingsfile, "STANDINGS", "opacity", str(opacity))
        appWriteCFGValue(settingsfile, "STANDINGS", "showBorder", str(showBorder))
        appWriteCFGValue(settingsfile, "STANDINGS", "showBadge", str(showBadge))
        appWriteCFGValue(settingsfile, "STANDINGS", "showTyres", str(showTyres))
        appWriteCFGValue(settingsfile, "STANDINGS", "showPitStops", str(showPitStops))
        appWriteCFGValue(settingsfile, "STANDINGS", "showDelta", str(showDelta))
        appWriteCFGValue(settingsfile, "STANDINGS", "colorAt", str(colorAt))
        appWriteCFGValue(settingsfile, "STANDINGS", "unit", str(unit))

        appWriteCFGValue(settingsfile, "STANDINGS", "driverDisplayedGlobal", str(driverDisplayedGlobal))
        appWriteCFGValue(settingsfile, "STANDINGS", "driverDisplayedClass", str(driverDisplayedClass))
        appWriteCFGValue(settingsfile, "STANDINGS", "driverDisplayedSameCar", str(driverDisplayedSameCar))

        appWriteCFGValue(settingsfile, "POSITIONS", "showLogoPos", str(showLogoPos))
        appWriteCFGValue(settingsfile, "POSITIONS", "showTitlePos", str(showTitlePos))
        appWriteCFGValue(settingsfile, "POSITIONS", "fontSizePos",  str(fontSizePos))
        appWriteCFGValue(settingsfile, "POSITIONS", "opacityPos",  str(opacityPos))
        appWriteCFGValue(settingsfile, "POSITIONS", "showBorderPos", str(showBorderPos))

        # configFile = open("apps/python/SubStanding/settings/settings.ini", 'w')
        # config.write(configFile)
        # configFile.close()

    except Exception as e:
        ac.log("SubStanding: Error in writeParameters: %s" % e)

def refreshAndWriteParameters():
    try:
        SubStandings["model"].refreshParameters(driverDisplayedSameCar)
        SubStandings["class"].refreshParameters(driverDisplayedClass)
        SubStandings["global"].refreshParameters(driverDisplayedGlobal)

        SubPositions["model"].refreshParameters()
        SubPositions["class"].refreshParameters()
        SubPositions["global"] .refreshParameters()

        configApp.updateView()
        writeParameters()

    except Exception as e:
        ac.log("SubStanding: Error in refreshAndWriteParameters: %s" % e)

def acUpdate(deltaT):
    try:
        global myDriver, lastUpdateTime, driversFullList, bdone

        lastUpdateTime += deltaT

        if lastUpdateTime < float(updateTime)/1000:
            return

        if debug:
            cpuUsage.startMesure()

        lastUpdateTime = 0

        driversList = []
        driversCount = info.static.numCars
        currentLapTime = ac.getCarState(0, acsys.CS.LapTime)

        modelStanding = []
        modelCount = 0
        myModelPos = 0

        classStanding = []
        classCount = 0
        myClassPos = 0

        globalStanding = []
        globalCount = 0
        myGlobalPos = 0

        # Get all needed infos
        for index in range(driversCount):
            driver = driversFullList[index]
            driver["connected"]   = ac.isConnected(index)
            driver["driverName"]  = ac.getDriverName(index)
            driver["driverNameShort"]  = ac.getDriverName(index)
            driver["carName"]     = ac.getCarName(index)
            driver["bestLap"]     = ac.getCarState(index, acsys.CS.BestLap)
            driver["lapTime"]     = ac.getCarState(index, acsys.CS.LapTime)
            driver["lapCount"]    = ac.getCarState(index, acsys.CS.LapCount)
            driver["speedMS"]     = ac.getCarState(index, acsys.CS.SpeedMS)
            driver["distance"]    = ac.getCarState(index, acsys.CS.NormalizedSplinePosition)
            # + ac.getCarState(index, acsys.CS.LapCount)
            driver["tyres"]       = ac.getCarTyreCompound(index)
            driver["isInPit"]     = ac.isCarInPit(index)
            #driver["realtimePos"] = ac.getCarLeaderboardPosition(index)
            driver["realtimePos"] = ac.getCarRealTimeLeaderboardPosition(index)
            # driver["realtimePos"] = ac.getCarState(index, acsys.CS.carRealTimeLeaderboardPosition)

            sp=driver["driverNameShort"].split('(')
            if len(sp)>1:
                # rest=sp[1]
                driver["driverNameShort"]=sp[0].strip()
                driver["driverName"]=driver["driverNameShort"]
                driver["aiStrength"]=sp[1].split(')')[0]

            sn = driver["driverNameShort"].split(' ')
            if len(sn)>1:
                if len(sn[1])<3:
                    driver["driverNameShort"] = str(sn[0][:6]) + '.' + str(sn[1][:2])
                else:
                    driver["driverNameShort"] = str(sn[0][:2]) + '.' + str(sn[1][:6])
            else:
                driver["driverNameShort"] = str(driver["driverNameShort"][:8])

            isInPitLine = ac.isCarInPitline(index)

            # Race
            if info.graphics.session == 2:
                # Trying to resolve the race start problem
                if driver["raceStartFlag"] and driver["lapCount"] == 0:
                    if driver["distance"] > 0.7:
                        driver["distance"] -= 1.0
                    elif driver["distance"] > 0.3:
                        driver["raceStartFlag"] = False

                    driver["pitStops"] = 0
                else:
                    driver["raceStartFlag"] = False

                # Pit entry
                if isInPitLine and not driver["isInPitLine"]:
                    driver["pitEntryFlag"] = True
                # Pit exit
                elif driver["pitEntryFlag"] and not isInPitLine and driver["isInPitLine"]:
                    driver["pitEntryFlag"] = False
                    driver["pitStops"] += 1
            else:
                driver["pitStops"] = 0
                driver["raceStartFlag"] = True

            driver["isInPitLine"] = isInPitLine

            if index == 0:
                myDriver = driver
                driversList.append(driver)
            elif driver["connected"] == 1:
                driversList.append(driver)

            driversFullList[index] = driver

        driversCount = len(driversList)

        # Sort the global standing by distance or best lap
        # Race
        if info.graphics.session == 2:
            # if driver["distance"] + driver["lapCount"]>
            #sort by distance
            driversList.sort(key=lambda tup: tup["distance"], reverse=True)
            # driversList.sort(key=lambda tup: (tup["distance"],tup["realtimePos"]), reverse=True)
            # driversList.sort(key=lambda tup: tup["realtimePos"], reverse=False)
            # while driversList.
            globalStanding = driversList
        # Practice, qualif
        else:
            # Only display drivers with a valid best lap
            for eachDriver in driversList:
                if eachDriver["bestLap"] == 0:
                    eachDriver["bestLap"] = 9999999
                globalStanding.append(eachDriver)

            #sort by bestLap
            globalStanding.sort(key=lambda tup: tup["bestLap"])

        # Extract sub-standings
        for eachDriver in globalStanding:
            if eachDriver["carName"] == myCarName:
                modelStanding.append(eachDriver)

            if eachDriver["carName"] in myClass:
                classStanding.append(eachDriver)

        # Update positions
        modelCount = len(modelStanding)
        if myDriver in modelStanding:
            myModelPos = modelStanding.index(myDriver) + 1
        else:
            myModelPos = 0

        classCount = len(classStanding)
        if myDriver in classStanding:
            myClassPos = classStanding.index(myDriver) + 1
        else:
            myClassPos = 0

        globalCount = len(globalStanding)
        if myDriver in globalStanding:
            myGlobalPos = globalStanding.index(myDriver) + 1
        else:
            myGlobalPos = 0

        # Update the views
        SubStandings["model"].updateView(modelStanding)
        SubStandings["class"].updateView(classStanding)
        SubStandings["global"].updateView(globalStanding)

        SubPositions["model"].updateView(myModelPos, modelCount)
        SubPositions["class"].updateView(myClassPos, classCount)
        SubPositions["global"].updateView(myGlobalPos, globalCount)

        configApp.updateView()

        if debug:
            cpuUsage.stopMesure()
            cpuUsage.updateView()

    except Exception as e:
        ac.log("SubStanding: Error in acUpdate: %s" % e)

def onRenderCallbackStandingModel(deltaT):
    try:
        SubStandings["model"].onRenderCallback(deltaT)
    except Exception as e:
        ac.log("SubStanding: Error in onRenderCallbackStandingModel: %s" % e)

def onRenderCallbackStandingClass(deltaT):
    try:
        SubStandings["class"].onRenderCallback(deltaT)
    except Exception as e:
        ac.log("SubStanding: Error in onRenderCallbackStandingClass: %s" % e)

def onRenderCallbackStandingGlobal(deltaT):
    try:
        SubStandings["global"].onRenderCallback(deltaT)
    except Exception as e:
        ac.log("SubStanding: Error in onRenderCallbackStandingGlobal: %s" % e)

def onRenderCallbackPositionModel(deltaT):
    try:
        SubPositions["model"].onRenderCallback(deltaT)
    except Exception as e:
        ac.log("SubStanding: Error in onRenderCallbackPositionModel: %s" % e)

def onRenderCallbackPositionClass(deltaT):
    try:
        SubPositions["class"].onRenderCallback(deltaT)
    except Exception as e:
        ac.log("SubStanding: Error in onRenderCallbackPositionClass: %s" % e)

def onRenderCallbackPositionGlobal(deltaT):
    try:
        SubPositions["global"].onRenderCallback(deltaT)
    except Exception as e:
        ac.log("SubStanding: Error in onRenderCallbackPositionGlobal: %s" % e)

def onRenderCallbackConfig(deltaT):
    try:
        configApp.onRenderCallback(deltaT)
    except Exception as e:
        ac.log("SubStanding: Error in onRenderCallbackConfig: %s" % e)

def isASCII(data):
    try:
        data.encode('ascii', 'ignore').decode('ascii').strip()
    except UnicodeDecodeError:
        return False
    else:
        return True

def find_str(s, char):
    index = 0
    if char in s:
        c = char[0]
        for ch in s:
            if ch == c:
                if s[index:index+len(char)] == char:
                    return index
            index += 1
    return -1

def appReadCFGValue(path, section, valname, default):
    try:
        result=default
        sect = ''
        if os.path.isfile(path):
            # ac.log(path)
            with io.open(path, 'r') as file:
                s = file.read()
            if not isASCII(s):
                with io.open(path, 'r', encoding='utf-8') as file: # read
                    s = file.read()
            ss = str(s).split('\n')
            for sss in ss: # for all lines
                sss=sss.strip()
                if find_str(sss, '--')!=0 and find_str(sss, ';')!=0 and find_str(sss, '#')!=0:
                    if ('[' in sss) and (']' in sss):
                        sect = sss
                    if (section=='' or sect=='['+section+']') and (valname in sss):
                        v = sss.replace(' = ', '=').split('=')
                        if len(v)>1:
                            if section=='' and '--' in v[1]:
                                w = v[1].split('--')
                                if len(w)>0:
                                    result = w[0].replace(' ', '')
                            else:
                                result = v[1]
                                # ac.log(valname + ' - ' + result)
    except:
        ac.log('SubStanding error: '+traceback.format_exc())
        return default
    return result

def appWriteCFGValue(path, section, valname, value):
    try:
        if os.path.isfile(path):
            # ini files with sections
            INIfilter = os.path.abspath(path)
            config = configparser.ConfigParser()  # encoding='utf8'
            config.optionxform = str
            with io.open(INIfilter, 'r', encoding='utf_8') as fp:
                config.read_file(fp)
            if not config.has_section(section):
                config.add_section(section)
            config.set(section, valname, value)
            with io.open(path, 'w', encoding='utf-8') as configfile:
                config.write(configfile, space_around_delimiters=False)
        else:
            ac.log("SubStanding : no file to write: " + path)
    except:
        ac.log("SubStanding error: " + traceback.format_exc())


class StandingWindow:

    def __init__(self, name, headerName, driverDisplayed):
        try:
            self.drivername_functions = {}
            self.headerName = headerName
            self.window = ac.newApp(name)

            self.positionLabel = []
            self.nameLabel = []
            self.tyresLabel = []
            self.pitLabel = []
            self.timeLabel = []
            self.carIdDisplayed = []
            self.currentDriverDisplayed = 1

            for index in range(maxDriverCount):
                self.positionLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.positionLabel[index], 'left')

                self.nameLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.nameLabel[index], 'left')

                #func = functools.partial(bool_change, sol_config_prefix, line_varname)
                #sol_config_functions[sol_config_prefix + line_varname] = func
                #ac.addOnCheckBoxChanged(sol_config_checkboxes[sol_config_prefix][line_varname],
                #                        sol_config_functions[sol_config_prefix + line_varname])

                func = functools.partial(self.onNameClick, index, self.nameLabel[index])  # declare a function
                self.drivername_functions[index] = func       # hold function in array
                ac.addOnClickedListener(self.nameLabel[index], self.drivername_functions[index]) # assign to label

                self.tyresLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.tyresLabel[index], 'center')

                self.pitLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.pitLabel[index], 'center')

                self.timeLabel.append(ac.addLabel(self.window, ""))
                ac.setFontAlignment(self.timeLabel[index], 'right')

                ac.setCustomFont(self.positionLabel[index], 'Segoe UI Black', 0 , 0)
                ac.setCustomFont(self.nameLabel[index], 'Segoe UI; Weight=Light', 0 , 0)
                ac.setCustomFont(self.tyresLabel[index], 'Segoe UI Black', 0 , 0)
                ac.setCustomFont(self.pitLabel[index], 'Segoe UI', 0 , 0)
                ac.setCustomFont(self.timeLabel[index], 'Segoe UI', 0 , 0)

                self.carIdDisplayed.append(-1)

            self.refreshParameters(driverDisplayed)
        except:
            ac.log("SubStanding Error: "  + traceback.format_exc())

    def onNameClick(self, index, labelno, *args):
        try:
            # ac.console( str(index) )          # id of clicked label
            # ac.console( ac.getText(labelno) ) # ac-id of clicked label
            # ac.console( str(self.carIdDisplayed[index] ) )
            # ac.setText(self.window, '\n\n' + str(args))
            ac.focusCar(self.carIdDisplayed[index])
        except:
            ac.log("SubStanding Error: "  + traceback.format_exc())

    def refreshParameters(self, driverDisplayed):
        self.driverDisplayed = driverDisplayed

        if showLogo:
            ac.setIconPosition(self.window, 0, 0)
        else:
            ac.setIconPosition(self.window, -10000, -10000)

        if showTitle:
            ac.setTitle(self.window, self.headerName)
        else:
            ac.setTitle(self.window, "")

        if showLogo or showTitle:
            self.firstSpacing = firstSpacing
        else:
            self.firstSpacing = 0

        self.widthPos   = fontSize*2
        if bUseMiniPreview:
            self.widthBadge = showBadge*(fontSize + spacing*4)
        else:
            self.widthBadge = showBadge*(fontSize + spacing)
        if shortNames:
            self.widthName  = fontSize*4
        else:
            self.widthName  = fontSize*10
        self.widthTyres = showTyres*fontSize*2
        self.widthPitStops = showPitStops*fontSize*2
        self.widthTime  = showDelta*fontSize*5
        self.width      = self.widthPos + self.widthBadge + self.widthName + self.widthTyres + self.widthPitStops + self.widthTime + 2*spacing
        height          = self.firstSpacing + (fontSize + spacing)*self.currentDriverDisplayed + spacing

        for index in range(maxDriverCount):
            ac.setFontSize(self.positionLabel[index], fontSize)
            ac.setPosition(self.positionLabel[index], spacing, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.positionLabel[index], self.widthPos/2+3, fontSize+spacing)

            ac.setFontSize(self.nameLabel[index], fontSize)
            ac.setPosition(self.nameLabel[index], spacing + self.widthPos + self.widthBadge, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.nameLabel[index], self.widthName, fontSize+spacing)
            # ac.setCustomFont(self.nameLabel[index], 'Segoe UI Light', 0 , 0)
            ac.setCustomFont(self.nameLabel[index], 'Segoe UI; Weight=Light', 0 , 0)

            ac.setFontSize(self.tyresLabel[index], fontSize)
            ac.setPosition(self.tyresLabel[index], spacing + self.widthPos + self.widthBadge + self.widthName, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.tyresLabel[index], self.widthTyres, fontSize+spacing)
            ac.setVisible(self.tyresLabel[index], showTyres)

            ac.setFontSize(self.pitLabel[index], fontSize)
            ac.setPosition(self.pitLabel[index], spacing + self.widthPos + self.widthBadge + self.widthName + self.widthTyres, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.pitLabel[index], self.widthPitStops, fontSize+spacing)
            ac.setVisible(self.pitLabel[index], showPitStops)

            ac.setFontSize(self.timeLabel[index], fontSize)
            ac.setPosition(self.timeLabel[index], spacing + self.widthPos + self.widthBadge + self.widthName + self.widthTyres + self.widthPitStops, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.timeLabel[index], self.widthTime, fontSize+spacing)
            ac.setVisible(self.timeLabel[index], showDelta)

        # Adjust window size, opacity and border
        ac.setSize(self.window, self.width, height)
        ac.setBackgroundOpacity(self.window, float(opacity)/100)
        if showBorder:
            ac.drawBorder(self.window, 1)
        else:
            ac.drawBorder(self.window, 0)

    def updateView(self, standing):
        global shortNames
        try:
            # Less drivers or equal than the max, display all of them
            if len(standing) <= self.driverDisplayed:
                driverIndex = 0
                lastIndex = len(standing) - 1
                self.currentDriverDisplayed = len(standing)
            # Only display "driverDisplayed" drivers
            else:
                myIndex = standing.index(myDriver)
                self.currentDriverDisplayed = self.driverDisplayed

                # We are on the top places of the standing, display from first
                if myIndex <= (self.driverDisplayed - 1)/2:
                    driverIndex = 0
                    lastIndex = self.driverDisplayed - 1
                # We are on the bottom places of the standing, display until last
                elif myIndex >= len(standing) - (self.driverDisplayed - 1)/2:
                    driverIndex = len(standing) - self.driverDisplayed
                    lastIndex = len(standing) - 1
                # We are on the middle of the standing
                else:
                    firstIndex = int(myIndex - (self.driverDisplayed - 1)/2)
                    if firstIndex < 0:
                        firstIndex = 0
                    lastIndex = firstIndex + self.driverDisplayed - 1
                    if lastIndex >= len(standing):
                        lastIndex = len(standing)-1
                    driverIndex = firstIndex

            # driverIndex

            for labelIndex in range(maxDriverCount):
                if driverIndex <= lastIndex:
                    ac.setVisible(self.positionLabel[labelIndex], 1)
                    ac.setVisible(self.nameLabel[labelIndex], 1)
                    ac.setVisible(self.tyresLabel[labelIndex], showTyres)
                    ac.setVisible(self.pitLabel[labelIndex], showPitStops)
                    ac.setVisible(self.timeLabel[labelIndex], showDelta)

                    # Refresh position and name
                    spos = int(standing[driverIndex]["realtimePos"])+1
                    if shortNames==True:
                        sname = standing[driverIndex]["driverNameShort"]
                    else:
                        sname = standing[driverIndex]["driverName"]
                    #ac.setText(self.positionLabel[labelIndex], "{0}.-".format(driverIndex+1))
                    ac.setText(self.positionLabel[labelIndex], str(spos) + ".")
                    ac.setText(self.nameLabel[labelIndex], sname)
                    self.carIdDisplayed[labelIndex] = standing[driverIndex]["carId"]

                    # Refresh time/distance
                    if info.graphics.session == 2:
                        # if Racing

                        if standing[driverIndex]["carId"] == 0:
                            # My car
                            if myDriver["lapCount"] > 1:
                                ac.setText(self.timeLabel[labelIndex],  "{0} laps".format(myDriver["lapCount"]))
                            else:
                                ac.setText(self.timeLabel[labelIndex],  "{0} lap".format(myDriver["lapCount"]))

                        else:
                            # Other car
                            # Delta string and color
                            # if standing[driverIndex]["lapCount"] < myDriver["lapCount"]:
                            deltaToLabel(self.timeLabel[labelIndex], (myDriver["lapCount"]+myDriver["distance"]) - (standing[driverIndex]["lapCount"]+standing[driverIndex]["distance"]), (standing[driverIndex]["speedMS"] + myDriver["speedMS"])/2)

                    else:
                        # Practice, qualif
                        ac.setText(self.timeLabel[labelIndex], timeToString(standing[driverIndex]["bestLap"]))
                        ac.setFontColor(self.timeLabel[labelIndex], 1, 1, 1, 1)

                    # My car
                    if standing[driverIndex]["carId"] == 0:
                        # Color my car in light blue
                        ac.setFontColor(self.positionLabel[labelIndex], 0, 1, 1, 1)
                        ac.setFontColor(self.nameLabel[labelIndex], 0, 1, 1, 1)
                        ac.setFontColor(self.tyresLabel[labelIndex], 0, 1, 1, 1)
                        ac.setFontColor(self.pitLabel[labelIndex], 0, 1, 1, 1)
                        ac.setFontColor(self.timeLabel[labelIndex], 0, 1, 1, 1)

                    # Other car
                    else:
                        # Position and name in blue for the teammates, pink for the rivals and white for the others
                        if standing[driverIndex]["driverName"] in teammates:
                            ac.setFontColor(self.positionLabel[labelIndex], 0, 1, 1, 1)
                            ac.setFontColor(self.nameLabel[labelIndex], 0, 1, 1, 1)
                            ac.setFontColor(self.tyresLabel[labelIndex], 0, 1, 1, 1)
                            ac.setFontColor(self.pitLabel[labelIndex], 0, 1, 1, 1)
                        elif standing[driverIndex]["driverName"] in rivals:
                            ac.setFontColor(self.positionLabel[labelIndex], 1, 0, 1, 1)
                            ac.setFontColor(self.nameLabel[labelIndex], 1, 0, 1, 1)
                            ac.setFontColor(self.tyresLabel[labelIndex], 1, 0, 1, 1)
                            ac.setFontColor(self.pitLabel[labelIndex], 1, 0, 1, 1)
                        else:
                            ac.setFontColor(self.positionLabel[labelIndex], 1, 1, 1, 1)
                            ac.setFontColor(self.nameLabel[labelIndex], 1, 1, 1, 1)
                            ac.setFontColor(self.tyresLabel[labelIndex], 1, 1, 1, 1)
                            ac.setFontColor(self.pitLabel[labelIndex], 1, 1, 1, 1)

                    ac.setText(self.tyresLabel[labelIndex], standing[driverIndex]["tyres"])

                    if standing[driverIndex]["isInPitLine"]:
                        ac.setText(self.pitLabel[labelIndex], "PIT")
                    elif info.graphics.session == 2:
                        ac.setText(self.pitLabel[labelIndex], "{0}p".format(standing[driverIndex]["pitStops"]))
                    else:
                        ac.setText(self.pitLabel[labelIndex], "")

                    driverIndex += 1
                else:
                    ac.setVisible(self.positionLabel[labelIndex], 0)
                    ac.setVisible(self.nameLabel[labelIndex], 0)
                    ac.setVisible(self.tyresLabel[labelIndex], 0)
                    ac.setVisible(self.pitLabel[labelIndex], 0)
                    ac.setVisible(self.timeLabel[labelIndex], 0)
                    ac.setText(self.timeLabel[labelIndex], "")
                    self.carIdDisplayed[labelIndex] = -1

            # Adjust window size, opacity and border
            ac.setSize(self.window, self.width, self.firstSpacing + (fontSize + spacing)*self.currentDriverDisplayed + spacing)

        except Exception as e:
            ac.log("SubStanding: Error in StandingWindow.updateView: %s" % e + '\n'+traceback.format_exc())

    def onRenderCallback(self, deltaT):
        global bUseMiniPreview
        for i in range(maxDriverCount):
            if self.carIdDisplayed[i] >= 0 and badge[self.carIdDisplayed[i]] > 0 and showBadge:
                ac.glColor4f(1,1,1,1)
                # x, y, w, h
                if bUseMiniPreview:
                    ac.glQuadTextured(self.widthPos-fontSize*0.5+2, self.firstSpacing + i*(fontSize+spacing),
                                      fontSize*2.0, fontSize*1.5, miniskin[self.carIdDisplayed[i]])
                else:
                    ac.glQuadTextured(self.widthPos, self.firstSpacing + i*(fontSize+spacing*1.2),
                                      fontSize,fontSize, badge[self.carIdDisplayed[i]])
            else:
                ac.glColor4f(1,1,1,0)
        ac.setBackgroundOpacity(self.window, float(opacity)/100)
        if showBorder:
            ac.drawBorder(self.window, 1)
        else:
            ac.drawBorder(self.window, 0)

def deltaToLabel(label, delta, speedAvg):
    if delta > 100:
        delta = 0
    distance = delta*trackLength
    if delta < 1 and delta > -1 and trackLength > 0:
        if unit == "metric":
            ac.setText(label, "{:+.0f}  m".format(distance))
        elif unit == "imperial":
            if distance < 1609:
                ac.setText(label, "{:+.0f}  ft".format(distance*3.28))
            else:
                ac.setText(label, "{:+.1f}  mi".format(distance/1609))
        elif unit == "time":
            if speedAvg < 15:
                speedAvg = 15
            deltaTime = -distance/speedAvg
            ac.setText(label, "{:+.1f}  s".format(deltaTime))
    elif delta < 2 and delta > -2:
        ac.setText(label, "{:+.1f}  lap".format(delta))
    else:
        ac.setText(label, "{:+.1f}  laps".format(delta))

    if trackLength > 0:
        distanceRelative = distance % trackLength
        if distance > 0 and distance < colorAt:
            # Green
            colorFactor = float(distance)/colorAt
            ac.setFontColor(label, colorFactor, 1, colorFactor, 1)
        elif distance < 0 and distance > (-colorAt):
            # Red
            colorFactor = float(distance)/(-colorAt)
            ac.setFontColor(label, 1, colorFactor, colorFactor, 1)
        elif distanceRelative > (trackLength - colorAt):
            # Blue
            colorFactor = float(distanceRelative - trackLength + colorAt)/colorAt
            ac.setFontColor(label, colorFactor, colorFactor, 1, 1)
        elif distanceRelative < colorAt:
            # Yellow
            colorFactor = float(distanceRelative)/colorAt
            ac.setFontColor(label, 1, 1, colorFactor, 1)
        else:
            # White
            ac.setFontColor(label, 1, 1, 1, 1)
    else:
        # White
        ac.setFontColor(label, 1, 1, 1, 1)

class PositionWindow:
    def __init__(self, name, headerName):
        self.headerName = headerName
        self.window = ac.newApp(name)
        ac.setTitle(self.window, headerName)

        self.label = ac.addLabel(self.window, "x/x")
        ac.setFontAlignment(self.label, 'center')

        self.refreshParameters()

    def refreshParameters(self):
        if showLogoPos:
            ac.setIconPosition(self.window, 0, 0)
        else:
            ac.setIconPosition(self.window, -10000, -10000)

        if showTitlePos:
            ac.setTitle(self.window, self.headerName)
        else:
            ac.setTitle(self.window, "")

        width  = fontSizePos*2 + 80
        height = fontSizePos + 50

        ac.setSize(self.window, width, height)

        ac.setPosition(self.label, 0, height/2 - fontSizePos/2)
        ac.setSize(self.label, width, fontSizePos)
        ac.setFontSize(self.label, fontSizePos)

    def updateView(self, myPos, carCount):
        ac.setText(self.label, "%d/%d" % (myPos, carCount))

    def onRenderCallback(self, deltaT):
        ac.setBackgroundOpacity(self.window, float(opacityPos)/100)
        if showBorderPos:
            ac.drawBorder(self.window, 1)
        else:
            ac.drawBorder(self.window, 0)

def timeToString(time):
    if time <= 0 or time == 9999999:
        return "-:--.---"
    else:
        return "{:d}:{:0>2d}.{:0>3d}".format(int(time/60000), int((time%60000)/1000), time%1000)



'''
Standings:
Show AC logo:   Yes     Change
Show title:     Yes     Change
Font size:      18      + -
Opacity:        50 %    + -
Show border:    Yes     Change
Show badges:    Yes     Change
-use skin  :    Yes     Change
Show pit stops: Yes     Change
Show deltas:    Yes     Change
Units:          Metric  Change
Color distance: 250 m   + -
Max drivers:
Global:         All     + -
Class:          All     + -
Same car:       All     + -

Positions:
Show AC logo:   Yes     Change
Show title:     Yes     Change
Font size:      18      + -
Opacity:        50 %    + -
Show border:    Yes     Change

Both:
Refresh every:  0.2 s   + -
'''
class SubStanding_config:

    def __init__(self, name, headerName, fontSize):
        self.window = ac.newApp(name)
        ac.setTitle(self.window, headerName)

        widthLeft       = fontSize*8
        widthCenter     = fontSize*5
        widthRight      = fontSize*5
        width           = widthLeft + widthCenter + widthRight + 2*spacing
        height          = firstSpacing + (fontSize*1.5 + spacing)*25

        ac.setSize(self.window, width, height)

        self.leftLabel = []
        self.centerLabel = []
        self.changeButton = []
        self.plusButton = []
        self.minusButton = []

        for index in range(28):
            self.leftLabel.append(ac.addLabel(self.window, ""))
            ac.setFontSize(self.leftLabel[index], fontSize)
            ac.setPosition(self.leftLabel[index], spacing, firstSpacing + index*(fontSize*1.25+spacing))
            ac.setSize(self.leftLabel[index], widthLeft, fontSize+spacing)
            ac.setFontAlignment(self.leftLabel[index], 'left')

            self.centerLabel.append(ac.addLabel(self.window, ""))
            ac.setFontSize(self.centerLabel[index], fontSize)
            ac.setPosition(self.centerLabel[index], spacing + widthLeft, firstSpacing + index*(fontSize*1.25+spacing))
            ac.setSize(self.centerLabel[index], widthCenter, fontSize+spacing)
            ac.setFontAlignment(self.centerLabel[index], 'left')

            self.changeButton.append(ac.addButton(self.window, "Change"))
            ac.setFontSize(self.changeButton[index], fontSize)
            ac.setPosition(self.changeButton[index], spacing + widthLeft + widthCenter, firstSpacing + index*(fontSize*1.25+spacing))
            ac.setSize(self.changeButton[index], fontSize*4, fontSize*1.5)

            self.plusButton.append(ac.addButton(self.window, "+"))
            ac.setFontSize(self.plusButton[index], fontSize)
            ac.setPosition(self.plusButton[index], spacing + widthLeft + widthCenter, firstSpacing + index*(fontSize*1.25+spacing))
            ac.setSize(self.plusButton[index], fontSize*1.5, fontSize*1.5)

            self.minusButton.append(ac.addButton(self.window, "-"))
            ac.setFontSize(self.minusButton[index], fontSize)
            ac.setPosition(self.minusButton[index], spacing + widthLeft + widthCenter + fontSize*2.5, firstSpacing + index*(fontSize*1.25+spacing))
            ac.setSize(self.minusButton[index], fontSize*1.5, fontSize*1.5)

        rowIndex = 0

        ac.setText(self.leftLabel[rowIndex], "Standings:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show AC logo:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleLogo)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showLogoId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show title:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleTitle)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showTitleId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Short Names:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleshortNames)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.shortNamesId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Font size:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], fontSizePlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], fontSizeMinus)
        self.fontSizeId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Opacity:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], opacityPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], opacityMinus)
        self.opacityId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show border:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleBorder)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showBorderId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show badges:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleBadge)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showBadgeId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "- use Skin:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleMiniPreview)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showMiniPreviewId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show tyres:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleTyres)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showTyresId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show pit stops:")
        ac.addOnClickedListener(self.changeButton[rowIndex], togglePitStops)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showPitStopsId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show delta:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleDelta)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showDeltaId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Units:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleUnits)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.unitsId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Color distance:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], colorAtPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], colorAtMinus)
        self.colorAtId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Max drivers:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "- Global:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], driverGlobalPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], driverGlobalMinus)
        self.driverGlobalId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "- Class:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], driverClassPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], driverClassMinus)
        self.driverClassId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "- Same car:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], driverSameCarPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], driverSameCarMinus)
        self.driverSameCarId = rowIndex

        rowIndex += 1

        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Positions:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show AC logo:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleLogoPos)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showLogoPosId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show title:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleTitlePos)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showTitlePosId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Font size:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], fontSizePosPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], fontSizePosMinus)
        self.fontSizePosId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Opacity:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], opacityPosPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], opacityPosMinus)
        self.opacityPosId = rowIndex

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Show border:")
        ac.addOnClickedListener(self.changeButton[rowIndex], toggleBorderPos)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)
        self.showBorderPosId = rowIndex

        rowIndex += 1

        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Global:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.setVisible(self.plusButton[rowIndex], 0)
        ac.setVisible(self.minusButton[rowIndex], 0)

        rowIndex += 1

        ac.setText(self.leftLabel[rowIndex], "Refresh every:")
        ac.setVisible(self.changeButton[rowIndex], 0)
        ac.addOnClickedListener(self.plusButton[rowIndex], refreshPlus)
        ac.addOnClickedListener(self.minusButton[rowIndex], refreshMinus)
        self.refreshId = rowIndex

    def updateView(self):
        ac.setText(self.centerLabel[self.showLogoId],  yesOrNo(showLogo))
        ac.setText(self.centerLabel[self.showTitleId], yesOrNo(showTitle))
        ac.setText(self.centerLabel[self.shortNamesId], yesOrNo(shortNames))
        ac.setText(self.centerLabel[self.fontSizeId],  str(fontSize))
        ac.setText(self.centerLabel[self.opacityId],   "{0} %".format(opacity))
        ac.setText(self.centerLabel[self.showBorderId], yesOrNo(showBorder))
        ac.setText(self.centerLabel[self.showTyresId], yesOrNo(showTyres))
        ac.setText(self.centerLabel[self.showPitStopsId], yesOrNo(showPitStops))
        ac.setText(self.centerLabel[self.showBadgeId], yesOrNo(showBadge))
        ac.setText(self.centerLabel[self.showMiniPreviewId], yesOrNo(bUseMiniPreview))
        ac.setText(self.centerLabel[self.showDeltaId], yesOrNo(showDelta))

        if unit == "metric":
            ac.setText(self.centerLabel[self.unitsId], "Metric")
            ac.setText(self.centerLabel[self.colorAtId], "{0} m".format(colorAt))
        elif unit == "imperial":
            ac.setText(self.centerLabel[self.unitsId], "Imperial")
            ac.setText(self.centerLabel[self.colorAtId], "{:.0f} ft".format(colorAt*3.28))
        elif unit == "time":
            ac.setText(self.centerLabel[self.unitsId], "Time (est)")
            ac.setText(self.centerLabel[self.colorAtId], "{0} m".format(colorAt))

        ac.setText(self.centerLabel[self.driverGlobalId], str(driverDisplayedGlobal))
        ac.setText(self.centerLabel[self.driverClassId], str(driverDisplayedClass))
        ac.setText(self.centerLabel[self.driverSameCarId], str(driverDisplayedSameCar))

        ac.setText(self.centerLabel[self.showLogoPosId],  yesOrNo(showLogoPos))
        ac.setText(self.centerLabel[self.showTitlePosId], yesOrNo(showTitlePos))
        ac.setText(self.centerLabel[self.fontSizePosId],  str(fontSizePos))
        ac.setText(self.centerLabel[self.opacityPosId],   "{0} %".format(opacityPos))
        ac.setText(self.centerLabel[self.showBorderPosId], yesOrNo(showBorderPos))

        if updateTime == 0:
            ac.setText(self.centerLabel[self.refreshId], "Min")
        elif updateTime == 50:
            ac.setText(self.centerLabel[self.refreshId], "0.05 s")
        else:
            ac.setText(self.centerLabel[self.refreshId], "{:.1f} s".format(float(updateTime)/1000))

    def onRenderCallback(self, deltaT):
        ac.setBackgroundOpacity(self.window, 1.0)

def yesOrNo(value):
    if value:
        return "Yes"
    else:
        return "No"

def toggleLogo(dummy, variable):
    global showLogo
    showLogo = not showLogo
    refreshAndWriteParameters()

def toggleTitle(dummy, variable):
    global showTitle
    showTitle = not showTitle
    refreshAndWriteParameters()

def toggleshortNames(dummy, variable):
    global shortNames
    shortNames = not shortNames
    refreshAndWriteParameters()

def fontSizePlus(dummy, variable):
    global fontSize
    fontSize += 1
    refreshAndWriteParameters()

def fontSizeMinus(dummy, variable):
    global fontSize
    if fontSize > 6:
        fontSize -= 1
    refreshAndWriteParameters()

def toggleBorder(dummy, variable):
    global showBorder
    showBorder = not showBorder
    refreshAndWriteParameters()

def toggleTyres(dummy, variable):
    global showTyres
    showTyres  = not showTyres
    refreshAndWriteParameters()

def togglePitStops(dummy, variable):
    global showPitStops
    showPitStops = not showPitStops
    refreshAndWriteParameters()

def toggleBadge(dummy, variable):
    global showBadge
    showBadge = not showBadge
    refreshAndWriteParameters()

def toggleMiniPreview(dummy, variable):
    global bUseMiniPreview
    bUseMiniPreview = not bUseMiniPreview
    refreshAndWriteParameters()

def opacityPlus(dummy, variable):
    global opacity
    if opacity < 100:
        opacity += 10
    refreshAndWriteParameters()

def opacityMinus(dummy, variable):
    global opacity
    if opacity >= 10:
        opacity -= 10
    refreshAndWriteParameters()

def toggleDelta(dummy, variable):
    global showDelta
    showDelta = not showDelta
    refreshAndWriteParameters()

def colorAtPlus(dummy, variable):
    global colorAt
    if colorAt < 2000:
        colorAt += 50
    refreshAndWriteParameters()

def colorAtMinus(dummy, variable):
    global colorAt
    if colorAt >= 100:
        colorAt -= 50
    refreshAndWriteParameters()

def toggleUnits(dummy, variable):
    global unit
    if unit == "metric":
        unit = "imperial"
    elif unit == "imperial":
        unit = "time"
    elif unit == "time":
        unit = "metric"
    refreshAndWriteParameters()

def driverGlobalPlus(dummy, variable):
    global driverDisplayedGlobal
    driverDisplayedGlobal += 2
    refreshAndWriteParameters()

def driverGlobalMinus(dummy, variable):
    global driverDisplayedGlobal
    if driverDisplayedGlobal > 3:
        driverDisplayedGlobal -= 2
    refreshAndWriteParameters()

def driverClassPlus(dummy, variable):
    global driverDisplayedClass
    driverDisplayedClass += 2
    refreshAndWriteParameters()

def driverClassMinus(dummy, variable):
    global driverDisplayedClass
    if driverDisplayedClass > 3:
        driverDisplayedClass -= 2
    refreshAndWriteParameters()

def driverSameCarPlus(dummy, variable):
    global driverDisplayedSameCar
    driverDisplayedSameCar += 2
    refreshAndWriteParameters()

def driverSameCarMinus(dummy, variable):
    global driverDisplayedSameCar
    if driverDisplayedSameCar > 3:
        driverDisplayedSameCar -= 2
    refreshAndWriteParameters()

def toggleLogoPos(dummy, variable):
    global showLogoPos
    showLogoPos = not showLogoPos
    refreshAndWriteParameters()

def toggleTitlePos(dummy, variable):
    global showTitlePos
    showTitlePos = not showTitlePos
    refreshAndWriteParameters()

def fontSizePosPlus(dummy, variable):
    global fontSizePos
    fontSizePos += 1
    refreshAndWriteParameters()

def fontSizePosMinus(dummy, variable):
    global fontSizePos
    if fontSizePos > 6:
        fontSizePos -= 1
    refreshAndWriteParameters()

def opacityPosPlus(dummy, variable):
    global opacityPos
    if opacityPos < 100:
        opacityPos += 10
    refreshAndWriteParameters()

def opacityPosMinus(dummy, variable):
    global opacityPos
    if opacityPos >= 10:
        opacityPos -= 10
    refreshAndWriteParameters()

def toggleBorderPos(dummy, variable):
    global showBorderPos
    showBorderPos = not showBorderPos
    refreshAndWriteParameters()

def refreshPlus(dummy, variable):
    global updateTime
    if updateTime == 0:
        updateTime = 50
    elif updateTime == 50:
        updateTime = 100
    elif updateTime < 1000:
        updateTime += 100
    else:
        updateTime = 1000
    refreshAndWriteParameters()

def refreshMinus(dummy, variable):
    global updateTime
    if updateTime == 50:
        updateTime = 0
    elif updateTime == 100:
        updateTime = 50
    elif updateTime > 100:
        updateTime -= 100
    else:
        updateTime = 0
    refreshAndWriteParameters()

class CpuUsage:
    def __init__(self, name, headerName, fontSize):
        self.name = name
        self.headerName = headerName
        self.window = ac.newApp(name)
        ac.setTitle(self.window, headerName)

        self.firstSpacing = firstSpacing

        widthLeft       = fontSize*6
        widthRight      = fontSize*5
        self.width      = widthLeft + widthRight + 2*spacing
        height          = self.firstSpacing + (fontSize + spacing)*5

        ac.setSize(self.window, self.width, height)

        self.leftLabel = []
        self.rightLabel = []

        for index in range(5):
            self.leftLabel.append(ac.addLabel(self.window, ""))
            ac.setFontSize(self.leftLabel[index], fontSize)
            ac.setPosition(self.leftLabel[index], spacing, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.leftLabel[index], widthLeft, fontSize+spacing)
            ac.setFontAlignment(self.leftLabel[index], 'left')

            self.rightLabel.append(ac.addLabel(self.window, ""))
            ac.setFontSize(self.rightLabel[index], fontSize)
            ac.setPosition(self.rightLabel[index], spacing + widthLeft, self.firstSpacing + index*(fontSize+spacing))
            ac.setSize(self.rightLabel[index], widthRight, fontSize+spacing)
            ac.setFontAlignment(self.rightLabel[index], 'right')

        ac.setText(self.leftLabel[0], "Avg time:")
        ac.setText(self.leftLabel[1], "Min time:")
        ac.setText(self.leftLabel[2], "Max time:")
        ac.setText(self.leftLabel[3], "Avg period:")
        ac.setText(self.leftLabel[4], "Avg load:")

        self.avg = 0
        self.min = 0
        self.max = 999999
        self.avgperiod = 0
        self.avgload = 0

        self.startTime = 0
        self.lastStartTime = 0
        self.delta = 0

    def startMesure(self):
        self.startTime = time.process_time()

        if self.lastStartTime == 0:
            self.lastStartTime = self.startTime

        self.lastPeriod = self.startTime - self.lastStartTime
        self.lastStartTime = self.startTime

    def stopMesure(self):
        self.delta = time.process_time() - self.startTime

        if self.avg == 0:
            self.avg = self.delta
            self.min = self.delta
            self.max = self.delta
        else:
            self.avg = self.avg*0.9 + self.delta*0.1
            if self.delta < self.min:
                self.min = self.delta
            elif self.delta > self.max:
                self.max = self.delta

        if self.avgperiod == 0:
            self.avgperiod = self.lastPeriod
        else:
            self.avgperiod = self.avgperiod*0.9 + self.lastPeriod*0.1

        if self.avgperiod == 0:
            self.avgload = 0
        else:
            self.avgload = self.avg/self.avgperiod*100

    def updateView(self):
        ac.setText(self.rightLabel[0], "{:.3f} ms".format(self.avg/1000))
        ac.setText(self.rightLabel[1], "{:.3f} ms".format(self.min/1000))
        ac.setText(self.rightLabel[2], "{:.3f} ms".format(self.max/1000))
        ac.setText(self.rightLabel[3], "{:.3f} ms".format(self.avgperiod/1000))
        ac.setText(self.rightLabel[4], "{:.3f} %".format(self.avgload))
